import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from featurestore.core.entities.permission import ManageablePermission, Permission
from featurestore.core.entities.permission_request import ManageablePermissionRequest, PermissionRequest


class AccessControlList:
    def __init__(self, stub):
        self.requests = AclRequests(stub)
        self.permissions = AclPermissions(stub)


class AclRequests:
    def __init__(self, stub):
        self.projects = AclProjectRequests(stub)
        self.feature_sets = AclFeatureSetsRequests(stub)


class AclPermissions:
    def __init__(self, stub):
        self.projects = AclProjectPermissions(stub)
        self.feature_sets = AclFeatureSetsPermissions(stub)


class AclProjectRequests:
    def __init__(self, stub):
        self._stub = stub

    def list(self):
        """List pending project permission requests.

        Returns:
            Generator of project permission requests

        Typical example:
            my_requests = client.acl.requests.projects.list()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#managing-permission-requests-from-other-users
        """
        request = pb.ListPermissionsPageRequest(filters=[pb.PermissionState.PENDING])
        return (
            PermissionRequest(self._stub, entry.permission, entry.project.name)
            for entry in paged_response_to_request(request, self._stub.ListProjectPermissionsPage)
        )

    def list_manageable(self):
        """List pending manageable project permission requests.

        Returns:
            Generator of manageable project permission requests

        Typical example:
            manageable_requests = client.acl.requests.projects.list_manageable()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#managing-permission-requests-from-other-users
        """
        request = pb.ListPermissionsPageRequest(filters=[pb.PermissionState.PENDING])
        return (
            ManageablePermissionRequest(self._stub, entry.permission, entry.project.name)
            for entry in paged_response_to_request(request, self._stub.ListManageableProjectPermissionsPage)
        )


class AclFeatureSetsRequests:
    def __init__(self, stub):
        self._stub = stub

    def list(self):
        """List pending feature set permission requests.

        Returns:
            Generator of feature set permission requests

        Typical example:
            my_requests = client.acl.requests.feature_sets.list()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#managing-permission-requests-from-other-users
        """
        request = pb.ListPermissionsPageRequest(filters=[pb.PermissionState.PENDING])
        return (
            PermissionRequest(self._stub, entry.permission, entry.feature_set.project_name)
            for entry in paged_response_to_request(request, self._stub.ListFeatureSetsPermissionsPage)
        )

    def list_manageable(self):
        """List pending manageable feature set permission requests.

        Returns:
            Generator of manageable feature set permission requests

        Typical example:
            manageable_requests = client.acl.requests.feature_sets.list_manageable()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#managing-permission-requests-from-other-users
        """
        request = pb.ListPermissionsPageRequest(filters=[pb.PermissionState.PENDING])
        return (
            ManageablePermissionRequest(self._stub, entry.permission, entry.feature_set.project_name)
            for entry in paged_response_to_request(request, self._stub.ListManageableFeatureSetsPermissionsPage)
        )


class AclProjectPermissions:
    def __init__(self, stub):
        self._stub = stub

    def list(self, filters=None):
        """List existing project permission requests.

        Args:
            filters: (list[PermissionState]) Filter includes the state of permission (either REJECTED or GRANTED).

        Returns:
            Generator of project permissions

        Typical example:
            filters = [PermissionState.REJECTED]
            my_requests = client.acl.requests.projects.list(filters)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#requesting-permissions-to-a-project
        """
        if filters is None:
            filters = [pb.PermissionState.GRANTED]
        request = pb.ListPermissionsPageRequest(filters=filters)
        return (
            Permission(self._stub, entry.permission, entry.project.name)
            for entry in paged_response_to_request(request, self._stub.ListProjectPermissionsPage)
        )

    def list_manageable(self, filters=None):
        """List pending manageable project permission requests.

        Args:
            filters: (list[PermissionState]) Filter includes the state of permission (either REJECTED or GRANTED).

        Returns:
            Generator of manageable project permissions

        Typical example:
            filters = [PermissionState.REJECTED]
            manageable_requests = client.acl.requests.projects.list_manageable(filters)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#managing-permission-requests-from-other-users
        """
        if filters is None:
            filters = [pb.PermissionState.GRANTED]
        request = pb.ListPermissionsPageRequest(filters=filters)
        return (
            ManageablePermission(self._stub, entry.permission, entry.project.name)
            for entry in paged_response_to_request(request, self._stub.ListManageableProjectPermissionsPage)
        )


class AclFeatureSetsPermissions:
    def __init__(self, stub):
        self._stub = stub

    def list(self, filters=None):
        """List pending feature set permission requests.

        Args:
            filters: (list[PermissionState]) Filter includes the state of permission (either REJECTED or GRANTED).

        Returns:
            Generator of feature set permissions

        Typical example:
            filters = [PermissionState.REJECTED]
            my_requests = client.acl.requests.feature_sets.list(filters)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#managing-feature-set-permissions
        """
        if filters is None:
            filters = [pb.PermissionState.GRANTED]
        request = pb.ListPermissionsPageRequest(filters=filters)
        return (
            Permission(self._stub, entry.permission, entry.feature_set.project_name)
            for entry in paged_response_to_request(request, self._stub.ListFeatureSetsPermissionsPage)
        )

    def list_manageable(self, filters=None):
        """List pending manageable feature set permission requests.

        Args:
            filters: (list[PermissionState]) Filter includes the state of permission (either REJECTED or GRANTED).

        Returns:
            Generator of manageable feature set permissions

        Typical example:
            filters = [PermissionState.REJECTED]
            manageable_requests = client.acl.requests.feature_sets.list_manageable(filters)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#managing-feature-set-permissions
        """
        if filters is None:
            filters = [pb.PermissionState.GRANTED]
        request = pb.ListPermissionsPageRequest(filters=filters)
        return (
            ManageablePermission(self._stub, entry.permission, entry.feature_set.project_name)
            for entry in paged_response_to_request(request, self._stub.ListManageableFeatureSetsPermissionsPage)
        )


def paged_response_to_request(request, core_call):
    while request:
        response = core_call(request)
        if response.next_page_token:
            request.page_token = response.next_page_token
        else:
            request = None
        for entry in response.entries:
            yield entry



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from ai.h2o.featurestore.api.v1.CoreService_pb2_grpc import CoreServiceStub

from ..entities.transformation_function import TransformationFunctionMetadata


class TransformationFunctions:
    def __init__(self, stub: CoreServiceStub):
        self._stub = stub

    def list(self):
        response: pb.ListTransformationFunctionsResponse = self._stub.ListTransformationFunctions(
            pb.ListTransformationFunctionsRequest()
        )
        return [
            TransformationFunctionMetadata(transformation_function)
            for transformation_function in response.transformation_functions
        ]

    def get(self, function_name: str):
        response = self._stub.GetTransformationFunction(
            pb.GetTransformationFunctionRequest(transformation_function_name=function_name)
        )
        return TransformationFunctionMetadata(response.transformation_function)



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
import ai.h2o.featurestore.api.v1.FeatureSetProtoApi_pb2 as FeatureSetApi

from ..schema import FeatureSchema, FeatureSchemaSpecialData, Schema


class FeatureSetSchema:
    def __init__(self, stub, feature_set):
        self._feature_set = feature_set
        self._stub = stub

    def get(self):
        """Get a schema of a feature set.

        Returns:
            Schema: A schema with feature names and data types.

            For example:

            id INT, text STRING, label DOUBLE, state STRING, date TIMESTAMP

        Typical example:
            fs.schema.get()
        """
        return Schema.create_from(self._feature_set)

    def is_compatible_with(self, new_schema, compare_data_types=True):
        """Compare a schema of a feature set with a schema of new data source.

        Args:
            new_schema: (Schema) A new schema to check compatibility with.
            compare_data_types: (bool) Object indicates whether data type needs to be compared or not.

        Returns:
            bool: A boolean describes whether compatible or not.

        Typical example:
            fs.schema.is_compatible_with(new_schema, compare_data_types=True)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#checking-schema-compatibility
        """
        request = pb.FeatureSetSchemaCompatibilityRequest()
        request.original_schema.extend(self.get()._to_proto_schema())
        request.new_schema.extend(new_schema._to_proto_schema())
        request.compare_data_types = compare_data_types
        response = self._stub.IsFeatureSetSchemaCompatible(request)
        return response.is_compatible

    def patch_from(self, new_schema, compare_data_types=True):
        """Patch a new schema with a schema of a feature set.

        Patch schema checks for matching features between the ‘new schema’ and the existing ‘fs.schema’.
        If there is a match, then the metadata such as special_data, description, etc are copied into the new_schema.

        Args:
            new_schema: (Schema) A new schema that needs to be patched.
            compare_data_types: (bool) Object indicates whether data type are to be compared while patching.

        Returns:
            Schema: A new schema after patches.

        Typical example:
            fs.schema.patch_from(new_schema, compare_data_types=True)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#patching-new-schema
        """
        request = pb.FeatureSetSchemaPatchRequest()
        request.original_schema.extend(self.get()._to_proto_schema())
        request.new_schema.extend(new_schema._to_proto_schema())
        request.compare_data_types = compare_data_types
        response = self._stub.FeatureSetSchemaPatch(request)
        return Schema(self._create_schema_from_proto(response.schema), True)

    @staticmethod
    def _create_schema_from_proto(schema):
        return [
            FeatureSchema(
                feature_schema.name,
                feature_schema.data_type,
                nested_features_schema=FeatureSetSchema._create_schema_from_proto(feature_schema.nested),
                special_data=FeatureSchemaSpecialData(
                    spi=feature_schema.special_data.spi,
                    pci=feature_schema.special_data.pci,
                    rpi=feature_schema.special_data.rpi,
                    demographic=feature_schema.special_data.demographic,
                    sensitive=feature_schema.special_data.sensitive,
                ),
                _feature_type=FeatureSetApi.FeatureType.Name(feature_schema.feature_type),
                description=feature_schema.description,
                classifiers=set(feature_schema.classifiers),
            )
            for feature_schema in schema
        ]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ai/h2o/featurestore/api/v1/TimeToLiveApi.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='ai/h2o/featurestore/api/v1/TimeToLiveApi.proto',
  package='ai.h2o.featurestore.api.v1',
  syntax='proto3',
  serialized_options=b'B\022TimeToLiveApiProto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n.ai/h2o/featurestore/api/v1/TimeToLiveApi.proto\x12\x1a\x61i.h2o.featurestore.api.v1\"L\n\x06Online\"B\n\x12TimeToLiveInterval\x12\x08\n\x04\x44\x41YS\x10\x00\x12\n\n\x06MONTHS\x10\x01\x12\t\n\x05HOURS\x10\x02\x12\x0b\n\x07MINUTES\x10\x03\"K\n\x07Offline\"@\n\x12TimeToLiveInterval\x12\t\n\x05YEARS\x10\x00\x12\x08\n\x04\x44\x41YS\x10\x01\x12\n\n\x06MONTHS\x10\x02\x12\t\n\x05HOURS\x10\x03\"\xdf\x01\n\nTimeToLive\x12\x13\n\x0bttl_offline\x18\x01 \x01(\r\x12T\n\x14ttl_offline_interval\x18\x02 \x01(\x0e\x32\x36.ai.h2o.featurestore.api.v1.Offline.TimeToLiveInterval\x12\x12\n\nttl_online\x18\x03 \x01(\r\x12R\n\x13ttl_online_interval\x18\x04 \x01(\x0e\x32\x35.ai.h2o.featurestore.api.v1.Online.TimeToLiveIntervalB\x14\x42\x12TimeToLiveApiProtob\x06proto3'
)



_ONLINE_TIMETOLIVEINTERVAL = _descriptor.EnumDescriptor(
  name='TimeToLiveInterval',
  full_name='ai.h2o.featurestore.api.v1.Online.TimeToLiveInterval',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DAYS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MONTHS', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HOURS', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MINUTES', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=88,
  serialized_end=154,
)
_sym_db.RegisterEnumDescriptor(_ONLINE_TIMETOLIVEINTERVAL)

_OFFLINE_TIMETOLIVEINTERVAL = _descriptor.EnumDescriptor(
  name='TimeToLiveInterval',
  full_name='ai.h2o.featurestore.api.v1.Offline.TimeToLiveInterval',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='YEARS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DAYS', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MONTHS', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HOURS', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=167,
  serialized_end=231,
)
_sym_db.RegisterEnumDescriptor(_OFFLINE_TIMETOLIVEINTERVAL)


_ONLINE = _descriptor.Descriptor(
  name='Online',
  full_name='ai.h2o.featurestore.api.v1.Online',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ONLINE_TIMETOLIVEINTERVAL,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=78,
  serialized_end=154,
)


_OFFLINE = _descriptor.Descriptor(
  name='Offline',
  full_name='ai.h2o.featurestore.api.v1.Offline',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _OFFLINE_TIMETOLIVEINTERVAL,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=156,
  serialized_end=231,
)


_TIMETOLIVE = _descriptor.Descriptor(
  name='TimeToLive',
  full_name='ai.h2o.featurestore.api.v1.TimeToLive',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ttl_offline', full_name='ai.h2o.featurestore.api.v1.TimeToLive.ttl_offline', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ttl_offline_interval', full_name='ai.h2o.featurestore.api.v1.TimeToLive.ttl_offline_interval', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ttl_online', full_name='ai.h2o.featurestore.api.v1.TimeToLive.ttl_online', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ttl_online_interval', full_name='ai.h2o.featurestore.api.v1.TimeToLive.ttl_online_interval', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=234,
  serialized_end=457,
)

_ONLINE_TIMETOLIVEINTERVAL.containing_type = _ONLINE
_OFFLINE_TIMETOLIVEINTERVAL.containing_type = _OFFLINE
_TIMETOLIVE.fields_by_name['ttl_offline_interval'].enum_type = _OFFLINE_TIMETOLIVEINTERVAL
_TIMETOLIVE.fields_by_name['ttl_online_interval'].enum_type = _ONLINE_TIMETOLIVEINTERVAL
DESCRIPTOR.message_types_by_name['Online'] = _ONLINE
DESCRIPTOR.message_types_by_name['Offline'] = _OFFLINE
DESCRIPTOR.message_types_by_name['TimeToLive'] = _TIMETOLIVE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Online = _reflection.GeneratedProtocolMessageType('Online', (_message.Message,), {
  'DESCRIPTOR' : _ONLINE,
  '__module__' : 'ai.h2o.featurestore.api.v1.TimeToLiveApi_pb2'
  # @@protoc_insertion_point(class_scope:ai.h2o.featurestore.api.v1.Online)
  })
_sym_db.RegisterMessage(Online)

Offline = _reflection.GeneratedProtocolMessageType('Offline', (_message.Message,), {
  'DESCRIPTOR' : _OFFLINE,
  '__module__' : 'ai.h2o.featurestore.api.v1.TimeToLiveApi_pb2'
  # @@protoc_insertion_point(class_scope:ai.h2o.featurestore.api.v1.Offline)
  })
_sym_db.RegisterMessage(Offline)

TimeToLive = _reflection.GeneratedProtocolMessageType('TimeToLive', (_message.Message,), {
  'DESCRIPTOR' : _TIMETOLIVE,
  '__module__' : 'ai.h2o.featurestore.api.v1.TimeToLiveApi_pb2'
  # @@protoc_insertion_point(class_scope:ai.h2o.featurestore.api.v1.TimeToLive)
  })
_sym_db.RegisterMessage(TimeToLive)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import datetime
import json
import logging
import time
from typing import Optional

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
import ai.h2o.featurestore.api.v1.FeatureSetProtoApi_pb2 as FeatureSetApi
import ai.h2o.featurestore.api.v1.OnlineApi_pb2 as OnlineApi
import ai.h2o.featurestore.api.v1.TimeToLiveApi_pb2 as TimeToLiveApi

from .. import interactive_console
from ..access_type import AccessType
from ..collections.ingest_history import IngestHistory
from ..collections.scheduled_tasks import ScheduledTasks
from ..credentials import CredentialsHelper
from ..data_source_wrappers import SparkDataFrame
from ..entities.backfill_option import BackfillOption
from ..entities.feature import Feature
from ..retrieve_holder import RetrieveHolder
from ..utils import Utils
from .feature_set_ref import FeatureSetRef
from .feature_set_schema import FeatureSetSchema
from .ingest_job import IngestJob
from .materialization_online_job import MaterializationOnlineJob
from .recommendation import Recommendation
from .user import User


class FeatureSet:
    def __init__(self, stub, feature_set):
        fs = pb.FeatureSet()
        fs.CopyFrom(feature_set)
        self._feature_set = fs
        self._stub = stub

    @property
    def id(self):
        return self._feature_set.id

    @property
    def project(self):
        return self._feature_set.project

    @property
    def feature_set_name(self):
        return self._feature_set.feature_set_name

    @property
    def version(self) -> str:
        return self._feature_set.version

    @property
    def major_version(self) -> int:
        return int(self.version.split(".")[0])

    @property
    def version_change(self):
        return self._feature_set.version_change

    @property
    def time_travel_column(self):
        return self._feature_set.time_travel_column

    @property
    def partition_by(self):
        return self._feature_set.partition_by

    @property
    def time_travel_column_format(self):
        return self._feature_set.time_travel_column_format

    @property
    def feature_set_type(self):
        return FeatureSetApi.FeatureSetType.Name(self._feature_set.feature_set_type)

    @feature_set_type.setter
    def feature_set_type(self, value):
        valid_values = FeatureSetApi.FeatureSetType.keys()
        if value.upper() in valid_values:
            update_request = pb.UpdateFeatureSetRequest(
                feature_set_id=self._feature_set.id,
                feature_set_version=self._feature_set.version,
                type=FeatureSetApi.FeatureSetType.Value(value.upper()),
                fields_to_update=[pb.FEATURE_SET_TYPE],
            )
            self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set
        else:
            raise Exception("Invalid feature set type. Supported values are: " + ", ".join(map(str, valid_values)))

    @property
    def description(self):
        return self._feature_set.description

    @description.setter
    def description(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            description=value,
            fields_to_update=[pb.FEATURE_SET_DESCRIPTION],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def author(self):
        return User(self._feature_set.author)

    @property
    def created_date_time(self):
        return Utils.timestamp_to_string(self._feature_set.created_date_time)

    @property
    def last_update_date_time(self):
        return Utils.timestamp_to_string(self._feature_set.last_update_date_time)

    @property
    def application_name(self):
        return self._feature_set.application_name

    @application_name.setter
    def application_name(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            application_name=value,
            fields_to_update=[pb.FEATURE_SET_APPLICATION_NAME],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def deprecated(self):
        return self._feature_set.deprecated

    @deprecated.setter
    def deprecated(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            deprecated=value,
            fields_to_update=[pb.FEATURE_SET_DEPRECATED],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def deprecated_date(self):
        return Utils.timestamp_to_string(self._feature_set.deprecated_date)

    @property
    def data_source_domains(self):
        return self._feature_set.data_source_domains

    @data_source_domains.setter
    def data_source_domains(self, value):
        if not isinstance(value, list):
            raise ValueError("data_source_domains accepts only list of strings as a value")

        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            fields_to_update=[pb.FEATURE_SET_DATA_SOURCE_DOMAINS],
        )
        update_request.data_source_domains.extend(value)
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def tags(self):
        return self._feature_set.tags

    @tags.setter
    def tags(self, value):
        if not isinstance(value, list):
            raise ValueError("tags accepts only list of strings as a value")

        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            fields_to_update=[pb.FEATURE_SET_TAGS],
        )
        update_request.tags.extend(value)
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def process_interval(self):
        return self._feature_set.process_interval

    @process_interval.setter
    def process_interval(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            process_interval=value,
            fields_to_update=[pb.FEATURE_SET_PROCESS_INTERVAL],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def process_interval_unit(self):
        return pb.ProcessIntervalUnit.Name(self._feature_set.process_interval_unit)

    @process_interval_unit.setter
    def process_interval_unit(self, value):
        valid_units = pb.ProcessIntervalUnit.keys()
        if value.upper() in valid_units:
            update_request = pb.UpdateFeatureSetRequest(
                feature_set_id=self._feature_set.id,
                feature_set_version=self._feature_set.version,
                process_interval_unit=pb.ProcessIntervalUnit.Value(value.upper()),
                fields_to_update=[pb.FEATURE_SET_PROCESS_INTERVAL_UNIT],
            )
            self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set
        else:
            raise Exception("Invalid process interval unit. Supported values are: " + ", ".join(map(str, valid_units)))

    @property
    def flow(self):
        return self._feature_set.flow

    @flow.setter
    def flow(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            flow=value,
            fields_to_update=[pb.FEATURE_SET_FLOW],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def features(self):
        return {
            feature.name: Feature(self._stub, self, feature, feature.name) for feature in self._feature_set.features
        }

    @property
    def primary_key(self):
        return self._feature_set.primary_key

    @property
    def statistics(self):
        return Statistics(self._feature_set)

    @property
    def time_to_live(self):
        return TimeToLive(self._stub, self)

    @property
    def special_data(self):
        return FeatureSetSpecialData(self._stub, self)

    @property
    def time_travel_scope(self):
        return FeatureSetScope(self._feature_set)

    @property
    def application_id(self):
        return self._feature_set.application_id

    @application_id.setter
    def application_id(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            application_id=value,
            fields_to_update=[pb.FEATURE_SET_APPLICATION_ID],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def feature_set_state(self):
        return self._feature_set.feature_set_state

    @feature_set_state.setter
    def feature_set_state(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            state=value,
            fields_to_update=[pb.FEATURE_SET_STATE],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def online(self):
        return Online(self._feature_set)

    @property
    def secret(self):
        return self._feature_set.secret

    @secret.setter
    def secret(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            secret=value,
            fields_to_update=[pb.FEATURE_SET_SECRET],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def custom_data(self):
        return self._feature_set.custom_data

    @property
    def feature_classifiers(self):
        return self._feature_set.feature_classifiers

    def is_derived(self):
        """Ensure its derivation from another feature set."""
        return self._feature_set.derived_from.HasField("transformation")

    def _reference(self) -> FeatureSetRef:
        return FeatureSetRef(self.id, self.major_version)

    @custom_data.setter
    def custom_data(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            custom_data=value,
            fields_to_update=[pb.FEATURE_SET_CUSTOM_DATA],
        )
        self._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    def create_new_version(
        self,
        schema=None,
        affected_features=None,
        reason="",
        primary_key=None,
        backfill_options: Optional[BackfillOption] = None,
    ):
        """Create a new version of feature set.

        Args:
            schema: (Schema) A feature set schema.
            affected_features: (list[str]) A list of feature names.
            reason: (str) A reason for creating new version. By default, an auto-generated message will be populated
              describing the features added/removed/modified.
            primary_key: (list[str]) A feature name / list of feature names.
            backfill_options: (BackfillOption) Object represents backfill. It is optional.

        Returns:
            FeatureSet: A feature set with new version.

        Typical example:
            new_fs = fs.create_new_version(schema=schema, reason="some message", primary_key=[])
            new_fs = fs.create_new_version(affected_features=["xyz"], reason="Computation of feature XYZ changed")
            new_fs = fs.create_new_version(schema=schema, reason="some message", backfill_options=backfill)

        Raises:
            ValueError: At least one of schema, affected_features or primary_key must be defined.

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_new_version.html
        """
        if schema is None and affected_features is None and primary_key is None:
            raise ValueError(
                "At least one of schema, affected_features or primary_key must be defined. "
                + "All values are supported as well"
            )
        request = pb.CreateNewFeatureSetVersionRequest()
        request.feature_set.CopyFrom(self._feature_set)
        request.reason = reason
        if schema:
            request.schema.extend(schema._to_proto_schema())
            if schema.is_derived():
                request.derived_from.CopyFrom(schema.derivation._to_proto())
        else:
            request.schema.extend(self.schema.get()._to_proto_schema())
        if affected_features:
            request.affected_features.extend(affected_features)
        if primary_key:
            request.primary_key.extend(primary_key)
        else:
            request.use_primary_key_from_previous_version = True
        if backfill_options:
            request.backfill_options.CopyFrom(backfill_options._to_proto(self._stub))
        response = self._stub.CreateNewFeatureSetVersion(request)
        if backfill_options:
            logging.info(
                f"Backfill started with job id: '{response.job.job_id}'. Please use this job id to track progress."
            )
        return FeatureSet(self._stub, response.feature_set)

    def refresh(self):
        """Refresh a feature set.

        This obtains the latest information of feature set.

        Returns:
            FeatureSet: A feature set with the latest information.
        """
        request = pb.GetFeatureSetsLastMinorForCurrentMajorRequest(
            feature_set_id=self._feature_set.id, feature_set_version=self._feature_set.version
        )
        self._feature_set = self._stub.GetFeatureSetsLastMinorForCurrentMajor(request).feature_set
        return self

    def delete(self, wait_for_completion=False):
        """Deletes feature set."""
        request = pb.DeleteFeatureSetRequest()
        request.feature_set.CopyFrom(self._feature_set)
        self._stub.DeleteFeatureSet(request)
        exists_request = pb.FeatureSetExistsRequest()
        exists_request.project_id = self._feature_set.project_id
        exists_request.feature_set_id = self._feature_set.id
        if wait_for_completion:
            while self._stub.FeatureSetExists(exists_request).exists:
                time.sleep(1)
                logging.debug(f"Waiting for feature set '{self._feature_set.feature_set_name}' deletion...")
        logging.info(f"Feature set '{self._feature_set.feature_set_name}' is deleted")

    def add_owners(self, user_emails):
        """Add additional owner/owners to a feature set.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            FeatureSet: An existing feature set with the latest information.

        Typical example:
            fs.add_owners(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#feature-set-permissions-api
        """
        return self._add_permissions(user_emails, pb.PermissionType.Owner)

    def add_editors(self, user_emails):
        """Add additional editor/editors to a feature set.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            FeatureSet: An existing feature set with the latest information.

        Typical example:
            fs.add_editors(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#feature-set-permissions-api
        """
        return self._add_permissions(user_emails, pb.PermissionType.Editor)

    def add_consumers(self, user_emails):
        """Add additional consumer/consumers to a feature set.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            FeatureSet: An existing feature set with the latest information.

        Typical example:
            fs.add_consumers(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#feature-set-permissions-api
        """
        return self._add_permissions(user_emails, pb.PermissionType.Consumer)

    def add_sensitive_consumers(self, user_emails):
        """Add additional sensitive consumer/consumers to a feature set.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            FeatureSet: An existing feature set with the latest information.

        Typical example:
            fs.add_sensitive_consumers(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#feature-set-permissions-api
        """
        return self._add_permissions(user_emails, pb.PermissionType.SensitiveConsumer)

    def remove_owners(self, user_emails):
        """Removes owner/owners from a feature set.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            FeatureSet: An existing feature set with the latest information.

        Typical example:
            fs.remove_owners(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#feature-set-permissions-api
        """
        return self._remove_permissions(user_emails, pb.PermissionType.Owner)

    def remove_editors(self, user_emails):
        """Remove editor/editors from a feature set.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            FeatureSet: An existing feature set with the latest information.

        Typical example:
            fs.remove_editors(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#feature-set-permissions-api
        """
        return self._remove_permissions(user_emails, pb.PermissionType.Editor)

    def remove_consumers(self, user_emails):
        """Remove consumer/consumers from a feature set.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            FeatureSet: An existing feature set with the latest information.

        Typical example:
            fs.remove_consumers(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#feature-set-permissions-api
        """
        return self._remove_permissions(user_emails, pb.PermissionType.Consumer)

    def remove_sensitive_consumers(self, user_emails):
        """Remove sensitive consumer/consumers from a feature set.

        Args:
            user_emails: (list[str]]) A collection of user emails.

        Returns:
            FeatureSet: An existing feature set with the latest information.

        Typical example:
            fs.remove_sensitive_consumers(["bob@h2o.ai", "alice@h2o.ai"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#feature-set-permissions-api
        """
        return self._remove_permissions(user_emails, pb.PermissionType.SensitiveConsumer)

    def get_active_jobs(self, job_type=pb.JobType.Unknown):
        """List running jobs belonging to a feature set.

        This returns a list of jobs that are currently processing for a specific feature set.
        It also retrieves a specific type of job.

        Args:
            job_type: (JobType) A job type.
              INGEST | RETRIEVE | EXTRACT_SCHEMA | REVERT_INGEST | MATERIALIZATION_ONLINE | COMPUTE_STATISTICS |
              COMPUTE_RECOMMENDATION_CLASSIFIERS | CREATE_ML_DATASET | BACKFILL

        Returns:
            list[BaseJob]: A collection of active jobs.

            For example:

            [Job(id=test123, type=Ingest, done=False, childJobIds=[]),
            Job(id=test456, type=Backfill, done=False, childJobIds=[])]

        Typical example:
            fs.get_active_jobs()
            fs.get_active_jobs(job_type=INGEST)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#feature-set-jobs-api
        """
        return self._get_jobs(True, job_type)

    def _get_jobs(self, active, job_type=pb.JobType.Unknown):
        from ..collections.jobs import Jobs  # Lazy import to avoid circular reference

        request = pb.ListJobsRequest(active=active)
        request.feature_set.CopyFrom(self._feature_set)
        request.job_type = job_type
        resp = self._stub.ListJobs(request)
        return [Jobs._create_job(self._stub, job_proto) for job_proto in resp.jobs]

    def _add_permissions(self, user_emails, permission):
        request = pb.FeatureSetPermissionRequest()
        request.feature_set.CopyFrom(self._feature_set)
        request.user_emails.extend(user_emails)
        request.permission = permission
        self._stub.AddFeatureSetPermission(request)
        return self

    def _remove_permissions(self, user_emails, permission):
        request = pb.FeatureSetPermissionRequest()
        request.feature_set.CopyFrom(self._feature_set)
        request.user_emails.extend(user_emails)
        request.permission = permission
        self._stub.RemoveFeatureSetPermission(request)
        return self

    def request_access(self, access_type, reason):
        """Request feature set permissions.

        Args:
            access_type: (PermissionType) A permission type.
                OWNER | EDITOR | CONSUMER | SENSITIVE_CONSUMER
            reason: (str) A reason for permission request.

        Returns:
            str: A permission id.

        Typical example:
            my_request_id = fs.request_access(AccessType.CONSUMER, "Preparing the best model")

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#managing-feature-set-permissions
        """
        request = pb.SubmitPendingPermissionRequest()
        request.resource_id = self._feature_set.id
        request.permission = AccessType.to_proto_permission(access_type)
        request.reason = reason
        response = self._stub.SubmitPendingFeatureSetPermission(request)
        return response.permission_id

    @property
    def current_permission(self):
        """Lists current feature set permissions."""
        request = pb.GetActivePermissionRequest(resource_id=self._feature_set.id)
        response = self._stub.GetActiveFeatureSetPermission(request)
        return AccessType.from_proto_active_permission(response.permission)

    def ingest_async(
        self,
        source,
        credentials=None,
    ):
        """Create an ingestion job for a feature set.

        Args:
            source: (CSVFile | CSVFolder | ParquetFile | ParquetFolder | JSONFile | JSONFolder |
              SnowflakeTable | SnowflakeCursor | JdbcTable | DeltaTable)
              A source location of supported data source.
            credentials: (AzureKeyCredentials | AzureSasCredentials | AzurePrincipalCredentials | S3Credentials |
              SnowflakeCredentials | TeradataCredentials | PostgresCredentials)
              To access the provided data source. Default is None.

        Returns:
            IngestJob: A job for ingestion.

            A job is created with unique id and type Ingest. For example:

            Job(id=<job_id>, type=Ingest, done=False, childJobIds=[])

        Typical example:
            fs.ingest_async(source, credentials=credentials)

        Raises:
            Exception: Manual ingest is not allowed on derived feature set.
        """
        if self.is_derived():
            raise Exception("Manual ingest is not allowed on derived feature set")
        from ..data_source_wrappers import get_raw_data_location

        if isinstance(source, SparkDataFrame):
            source._write_to_cache(self._stub)
            data_source = source._get_cache_location()
        else:
            data_source = get_raw_data_location(source)
        request = pb.StartIngestJobRequest()
        request.feature_set.CopyFrom(self._feature_set)
        request.data_source.CopyFrom(data_source)
        CredentialsHelper.set_credentials(request, data_source, credentials)
        job_id = self._stub.StartIngestJob(request)
        return IngestJob(self._stub, job_id)

    @interactive_console.record_stats
    def ingest(
        self,
        source,
        credentials=None,
    ):
        """Ingest data into the Feature Store (Offline ingestion).

        Args:
            source: (CSVFile | CSVFolder | ParquetFile | ParquetFolder | JSONFile | JSONFolder |
              SnowflakeTable | SnowflakeCursor | JdbcTable | DeltaTable)
              A source location of supported data source.
            credentials: (AzureKeyCredentials | AzureSasCredentials | AzurePrincipalCredentials | S3Credentials |
              SnowflakeCredentials | TeradataCredentials | PostgresCredentials)
              To access the provided data source. Default is None.

        Typical example:
            fs.ingest(source, credentials=credentials)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#offline-ingestion
        """
        job = self.ingest_async(source, credentials)
        result = job.wait_for_result()
        self._feature_set = result._get_feature_set()
        return result

    def materialize_online_async(self):
        """Create a job for feature set online materialization.

        Returns:
            MaterializationOnlineJob: A job for online materialization.

            A job is created with a unique id and type MaterializationOnline. For example:

            Job(id=<job_id>, type=MaterializationOnline, done=False, childJobIds=[])

        Typical example:
            future = feature_set.materialize_online_async()
        """
        request = pb.StartMaterializationOnlineRequest()
        request.feature_set.CopyFrom(self._feature_set)
        job_id = self._stub.StartMaterializationOnlineJob(request)
        return MaterializationOnlineJob(self._stub, job_id)

    @interactive_console.record_stats
    def materialize_online(self):
        """Materialize a feature set in the offline store to online.

        This pushes existing data from offline Feature store into online.

        Typical example:
            feature_set.materialize_online()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#offline-to-online-api
        """
        job = self.materialize_online_async()
        return job.wait_for_result()

    def ingest_online(self, rows):
        """Ingest data into the online Feature Store.

        Args:
            rows: (list[str] | str) Rows can be a single JSON row (string) or an array of JSON strings.

        Typical example:
            feature_set.ingest_online('{"id": 1, "label": "Carl"}')

        Raises:
            Exception: Manual ingest online is not allowed on derived feature set.

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#online-ingestion
        """
        if self.is_derived():
            raise Exception("Manual ingest online is not allowed on derived feature set")

        if isinstance(rows, list):
            row_list = rows
        else:
            row_list = [rows]
        request = OnlineApi.OnlineIngestRequest(
            feature_set_id=self._feature_set.id, feature_set_version=self._feature_set.version, rows=row_list
        )
        self._stub.OnlineIngest(request)

    def retrieve_online(self, *key) -> dict:
        """Retrieve data from the online Feature Store.

        Args:
            key: (Any) A specific primary key value for which the entry is obtained.

        Returns:
            dict: A dictionary of specific instance (JSON row).

            For example:

            {'id': '01', 'department': 'Engineering', 'name': 'Test'}

        Typical example:
            json = feature_set.retrieve_online(key)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#retrieving-from-online
        """
        request = OnlineApi.OnlineRetrieveRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
            key=map(lambda x: str(x), key),
        )
        json_row = self._stub.OnlineRetrieve(request).row
        return json.loads(json_row)

    def retrieve(self, start_date_time=None, end_date_time=None):
        """Retrieve data.

        This retrieves only a specific range of ingested data. If the parameters are empty,
        all data are fetched.

        Args:
            start_date_time: (str) A start date and time of the range of ingested data.
            end_date_time: (str) A end date and time of the range of ingested data.

        Returns:
            RetrieveHolder: Returns a link as output for reference.

        Typical example:
            ref = fs.retrieve(start_date_time="2023-01-01 00:00:00", end_date_time="2023-01-02 00:00:00")

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#retrieving-api
        """
        return RetrieveHolder(self._stub, self._feature_set, start_date_time, end_date_time, "")

    def list_versions(self):
        """List all versions of a feature set.

        This shows all versions of a feature set (the current and previous ones).

        Returns:
            list[VersionDescription]: A list of versions and its details.

            For example:

            [{
              "version": "1.0",
              "versionChange": "Feature Set Created",
              "createdDateTime": "2023-01-01T00:00:00.000Z"
            }]

        Typical example:
            fs.list_versions()
        """
        request = pb.ListFeatureSetsVersionRequest()
        request.feature_set.CopyFrom(self._feature_set)
        response = self._stub.ListFeatureSetVersions(request)
        return [VersionDescription(version) for version in response.versions]

    @property
    def schema(self):
        return FeatureSetSchema(self._stub, self)

    def ingest_history(self):
        """Create an ingest history containing all ingestion.

        Returns:
            IngestHistory: An object of IngestHistory class.

        Typical example:
            history = my_feature_set.ingest_history()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/ingest_history_api.html#getting-the-ingestion-history
        """
        return IngestHistory(self._stub, self._feature_set)

    def get_recommendations(self):
        """Get feature set recommendations.

        Returns:
            list[Recommendation]: A list of recommendations.

        Typical example:
            fs.get_recommendations()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#getting-recommendations
        """
        request = pb.GetRecommendationRequest()
        request.feature_set.CopyFrom(self._feature_set)
        response = self._stub.GetRecommendations(request)
        return [Recommendation(self._stub, self, item) for item in response.matches]

    def schedule_ingest(self, name: str, source, schedule: str, description: str = "", credentials=None):
        """Create a new scheduled task.

        Args:
            name: (str) A task name.
            source: (CSVFile | CSVFolder | ParquetFile | ParquetFolder | JSONFile | JSONFolder |
              SnowflakeTable | SnowflakeCursor | JdbcTable | DeltaTable)
              Source location of a supported data source.
            schedule: (str) Schedule should be in cron format.
            description: (str) Description about a task.
            credentials: (AzureKeyCredentials | AzureSasCredentials | AzurePrincipalCredentials | S3Credentials |
              SnowflakeCredentials | TeradataCredentials | PostgresCredentials)
              To access the provided data source. Default is None.

        Returns:
            ScheduledTask: A new scheduled task.

        Typical example:
            fs.schedule_ingest("task_name", source, schedule = "0 2 * * *", description = "", credentials = None)

        Raises:
            Exception: Scheduling Ingest with SparkDataFrame is not supported.

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_schedule.html#schedule-a-new-task
        """
        from ..data_source_wrappers import get_raw_data_location

        if isinstance(source, SparkDataFrame):
            raise Exception("Scheduling Ingest with SparkDataFrame is not supported.")
        else:
            data_source = get_raw_data_location(source)

        client_timezone = datetime.datetime.utcnow().astimezone().tzname()

        request = pb.ScheduleTaskRequest(
            name=name,
            description=description,
            feature_set_id=self._feature_set.id,
            project_id=self._feature_set.project_id,
            source=data_source,
            schedule=schedule,
            feature_set_version=self._feature_set.version,
            crone_time_zone=client_timezone,
        )
        CredentialsHelper.set_credentials(request, data_source, credentials)
        scheduled_task = ScheduledTasks(self._stub, self)
        return scheduled_task.create_ingest_task(request)

    @property
    def schedule(self):
        return ScheduledTasks(self._stub, self)

    def get_preview(self):
        """Preview the returned data.

        This previews up to a maximum of 100 rows and 50 features.

        Returns:
            list[dict]: A list of dictionary which contains JSON rows.

        Typical example:
            fs.get_preview()
        """
        request = pb.GetFeatureSetPreviewRequest(
            feature_set_id=self._feature_set.id,
            feature_set_version=self._feature_set.version,
        )
        response = self._stub.GetFeatureSetPreview(request)
        if response.preview_url:
            json_response = Utils.fetch_preview_as_json_array(response.preview_url)
            return json_response
        else:
            return []

    def ingest_lazy(self, source, credentials=None):
        """Ingest data lazy.

        Args:
            source: (CSVFile | CSVFolder | ParquetFile | ParquetFolder | JSONFile | JSONFolder |
              SnowflakeTable | SnowflakeCursor | JdbcTable | DeltaTable)
              A source location of supported data source.
            credentials: (AzureKeyCredentials | AzureSasCredentials | AzurePrincipalCredentials | S3Credentials |
              SnowflakeCredentials | TeradataCredentials | PostgresCredentials)
              To access the provided data source. Default is None.

        Returns:
            ScheduledTask: A new scheduled task.

        Typical example:
            fs.ingest_lazy(source)

        Raises:
            Exception: Lazy Ingest with SparkDataFrame is not supported.
        """
        from ..data_source_wrappers import get_raw_data_location

        if isinstance(source, SparkDataFrame):
            raise Exception("Lazy Ingest with SparkDataFrame is not supported.")
        else:
            data_source = get_raw_data_location(source)
        request = pb.ScheduleTaskRequest(
            feature_set_id=self._feature_set.id,
            project_id=self._feature_set.project_id,
            source=data_source,
            feature_set_version=self._feature_set.version,
        )
        CredentialsHelper.set_credentials(request, data_source, credentials)
        scheduled_task = ScheduledTasks(self._stub, self)
        scheduled_task.create_lazy_ingest_task(request)
        logging.info(f"Lazy ingest scheduled for feature set {self._feature_set.id}")

    @property
    def owners(self):
        request = pb.GetOwnersRequest(resource_id=self._feature_set.id)
        response = self._stub.GetFeatureSetOwners(request)
        return [User(owner) for owner in response.owners]

    def _start_preview_job(self):
        request = pb.PreviewRequest()
        request.feature_set.CopyFrom(self._feature_set)
        return self._stub.StartPreviewJob(request)

    def _get_job(self, job_id):
        return self._stub.GetJob(job_id)

    def __repr__(self):
        return Utils.pretty_print_proto(self._feature_set)

    def __str__(self):
        return (
            f"Feature set name    : {self.feature_set_name} \n"
            f"Description         : {self.description} \n"
            f"Version             : {self.version} \n"
            f"Secret              : {self.secret} \n"
            f"Author                \n{Utils.output_indent_spacing(str(self.author), '      ')}"
            f"Project name        : {self.project} \n"
            f"Primary key         : {self.primary_key} \n"
            f"Feature set type    : {self.feature_set_type} \n"
            f"Created             : {self.created_date_time} \n"
            f"Last updated        : {self.last_update_date_time} \n"
            f"Sensitive           : {self.special_data._fs._feature_set.special_data.sensitive} \n"
            f"Time travel column  : {self.time_travel_column} \n"
            f"Features            : {self._custom_feature_fields()} \n"
            f"Tags                : {self.tags} \n"
            f"Partition by        : {self.partition_by} \n"
            f"Feature classifiers : {self.feature_classifiers} \n"
        )

    def _custom_feature_fields(self):
        tmp_dict = dict()
        for feature in self.features:
            tmp_dict.update({feature: self.features.get(feature).data_type})
        return json.dumps(tmp_dict, indent=5)


class VersionDescription:
    def __init__(self, version_description):
        self._version_description = version_description

    def __repr__(self):
        return Utils.pretty_print_proto(self._version_description)

    def __str__(self):
        return (
            f"Version           : {Utils.proto_to_dict(self._version_description).get('version')} \n"
            f"Version change    : {Utils.proto_to_dict(self._version_description).get('version_change')} \n"
            f"Created           : {Utils.proto_to_dict(self._version_description).get('created_date_time')} \n"
        )


class TimeToLive:
    def __init__(self, stub, feature_set):
        self._stub = stub
        self._fs = feature_set

    @property
    def ttl_offline(self):
        return self._fs._feature_set.time_to_live.ttl_offline

    @ttl_offline.setter
    def ttl_offline(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._fs._feature_set.id,
            feature_set_version=self._fs._feature_set.version,
            time_to_live_offline_interval=value,
            fields_to_update=[pb.FEATURE_SET_TTL_OFFLINE],
        )
        self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def ttl_offline_interval(self):
        return TimeToLiveApi.Offline.TimeToLiveInterval.Name(self._fs._feature_set.time_to_live.ttl_offline_interval)

    @ttl_offline_interval.setter
    def ttl_offline_interval(self, value):
        valid_units = TimeToLiveApi.Offline.TimeToLiveInterval.keys()
        if value.upper() in valid_units:
            update_request = pb.UpdateFeatureSetRequest(
                feature_set_id=self._fs._feature_set.id,
                feature_set_version=self._fs._feature_set.version,
                time_to_live_offline_interval_unit=TimeToLiveApi.Offline.TimeToLiveInterval.Value(value.upper()),
                fields_to_update=[pb.FEATURE_SET_TTL_OFFLINE_INTERVAL],
            )
            self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set
        else:
            raise Exception(
                "Invalid offline time to live interval unit. Supported values are: " + ", ".join(map(str, valid_units))
            )

    @property
    def ttl_online(self):
        return self._fs._feature_set.time_to_live.ttl_online

    @ttl_online.setter
    def ttl_online(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._fs._feature_set.id,
            feature_set_version=self._fs._feature_set.version,
            time_to_live_online_interval=value,
            fields_to_update=[pb.FEATURE_SET_TTL_ONLINE],
        )
        self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def ttl_online_interval(self):
        return TimeToLiveApi.Online.TimeToLiveInterval.Name(self._fs._feature_set.time_to_live.ttl_online_interval)

    @ttl_online_interval.setter
    def ttl_online_interval(self, value):
        valid_units = TimeToLiveApi.Online.TimeToLiveInterval.keys()
        if value.upper() in valid_units:
            update_request = pb.UpdateFeatureSetRequest(
                feature_set_id=self._fs._feature_set.id,
                feature_set_version=self._fs._feature_set.version,
                time_to_live_online_interval_unit=TimeToLiveApi.Online.TimeToLiveInterval.Value(value.upper()),
                fields_to_update=[pb.FEATURE_SET_TTL_ONLINE_INTERVAL],
            )
            self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set
        else:
            raise Exception(
                "Invalid online time to live interval unit. Supported values are: " + ", ".join(map(str, valid_units))
            )

    def __repr__(self):
        return Utils.pretty_print_proto(self._fs._feature_set.time_to_live)

    def __str__(self):
        return (
            f"ttl offline           : {self.ttl_offline} \n"
            f"ttl online            : {self.ttl_online} \n"
            f"ttl offline interval  : {self.ttl_offline_interval} \n"
            f"ttl online interval   : {self.ttl_online_interval} \n"
        )


class FeatureSetScope:
    def __init__(self, feature_set):
        self._feature_set = feature_set
        self._scope = self._feature_set.time_travel_scope

    @property
    def start_date_time(self):
        return Utils.timestamp_to_string(self._scope.start_date_time)

    @property
    def end_date_time(self):
        return Utils.timestamp_to_string(self._scope.end_date_time)

    def __repr__(self):
        return Utils.pretty_print_proto(self._scope)

    def __str__(self):
        return f"Start date & time : {self.start_date_time} \n" f"End date & time   : {self.end_date_time} \n"


class FeatureSetSpecialData:
    def __init__(self, stub, feature_set):
        self._stub = stub
        self._fs = feature_set

    @property
    def spi(self):
        return self._fs._feature_set.special_data.spi

    @property
    def pci(self):
        return self._fs._feature_set.special_data.pci

    @property
    def rpi(self):
        return self._fs._feature_set.special_data.rpi

    @property
    def demographic(self):
        return self._fs._feature_set.special_data.demographic

    @property
    def legal(self):
        return Legal(self._stub, self._fs)

    def __repr__(self):
        return Utils.pretty_print_proto(self._fs._feature_set.special_data)

    def __str__(self):
        return (
            f"legal           \n{Utils.output_indent_spacing(str(self.legal), '    ')}"
            f"spi           : {self.spi} \n"
            f"pci           : {self.pci} \n"
            f"rpi           : {self.rpi} \n"
            f"demographic   : {self.demographic} \n"
            f"sensitive     : {Utils.proto_to_dict(self._fs._feature_set.special_data).get('sensitive')} \n"
        )


class Statistics:
    def __init__(self, feature_set):
        self._feature_set = feature_set
        self._statistics = self._feature_set.statistics

    @property
    def data_latency(self):
        return self._statistics.data_latency

    def __repr__(self):
        return Utils.pretty_print_proto(self._statistics)

    def __str__(self):
        return f"Data latency  : {self.data_latency} \n"


class Legal:
    def __init__(self, stub, feature_set):
        self._stub = stub
        self._fs = feature_set

    @property
    def approved(self):
        return self._fs._feature_set.special_data.legal.approved

    @approved.setter
    def approved(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._fs._feature_set.id,
            feature_set_version=self._fs._feature_set.version,
            legal_approved=value,
            fields_to_update=[pb.FEATURE_SET_SPECIAL_DATA_LEGAL_APPROVED],
        )
        self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    @property
    def approved_date(self):
        return Utils.timestamp_to_string(self._fs._feature_set.special_data.legal.approved_date)

    @property
    def notes(self):
        return self._fs._feature_set.special_data.legal.notes

    @notes.setter
    def notes(self, value):
        update_request = pb.UpdateFeatureSetRequest(
            feature_set_id=self._fs._feature_set.id,
            feature_set_version=self._fs._feature_set.version,
            legal_approved_notes=value,
            fields_to_update=[pb.FEATURE_SET_SPECIAL_DATA_LEGAL_NOTES],
        )
        self._fs._feature_set = self._stub.UpdateFeatureSet(update_request).updated_feature_set

    def __repr__(self):
        return Utils.pretty_print_proto(self._fs._feature_set.special_data.legal)

    def __str__(self):
        return f"Approved  : {self.approved} \n" f"Notes     : {self.notes} \n"


class Online:
    def __init__(self, feature_set):
        self._feature_set = feature_set
        self._online = self._feature_set.online

    @property
    def online_namespace(self):
        return self._online.online_namespace

    @property
    def connection_type(self):
        return self._online.connection_type

    @property
    def topic(self):
        return self._online.topic

    def __repr__(self):
        return Utils.pretty_print_proto(self._online)

    def __str__(self):
        return (
            f"Online namespace  : {self.online_namespace} \n"
            f"Connection type   : {self.connection_type} \n"
            f"Topic             : {self.topic} \n"
        )



import datetime
from typing import Optional, Tuple

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from ai.h2o.featurestore.api.v1.CoreService_pb2_grpc import CoreServiceStub

from .. import interactive_console
from ..entities.create_ml_dataset_job import CreateMLDatasetJob
from ..entities.ml_dataset import MLDataset
from ..utils import Utils


class MLDatasets:
    def __init__(self, stub: CoreServiceStub, feature_view):
        self._feature_view = feature_view
        self._stub = stub

    @interactive_console.record_stats
    def create(
        self,
        name: str,
        description: str = "",
        start_date_time: Optional[datetime.datetime] = None,
        end_date_time: Optional[datetime.datetime] = None,
    ):
        """Create a machine learning dataset (MLDataset).

        Args:
            name: (str) Title of a new MLDataset.
            description: (str) Description about a new MLDataset.
            start_date_time: (datetime.datetime) Object represents start time for setting the start line
              of range in ingestion data.
            end_date_time: (datetime.datetime) Object represents end time for setting the last line
              of range in ingestion data.

        Returns:
            MLDataset: Machine Learning Dataset

        Typical example:
            ml_dataset = my_feature_view.ml_datasets.create("name", start_date_time=None, end_date_time=None)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_view_api.html#creating-a-mldataset
        """
        job, ml_dataset = self.create_async(name, description, start_date_time, end_date_time)
        return job.wait_for_result()

    def create_async(
        self,
        name: str,
        description: str = "",
        start_date_time: Optional[datetime.datetime] = None,
        end_date_time: Optional[datetime.datetime] = None,
    ) -> Tuple[CreateMLDatasetJob, MLDataset]:
        """Create a job for machine learning dataset (MLDataset) creation.

        Args:
            name: (str) Title for a new MLDataset.
            description: (str) Description about a new MLDataset.
            start_date_time: (datetime.datetime) Object represents start time for setting the start line
              of range in ingestion data.
            end_date_time: (datetime.datetime) Object represents end time for setting the last line
              of range in ingestion data.

        Returns:
            Tuple[CreateMLDatasetJob, MLDataset]

            CreateMLDatasetJob: Represents a job

        Typical example:
            my_feature_view.ml_datasets.create_async("name", start_date_time=None, end_date_time=None)
        """
        request = pb.CreateMLDatasetRequest(
            name=name,
            description=description,
            feature_view_id=self._feature_view.id,
            feature_view_version=self._feature_view.version,
            start_date_time=Utils.date_time_to_proto_timestamp(start_date_time),
            end_date_time=Utils.date_time_to_proto_timestamp(end_date_time),
        )

        response = self._stub.CreateMLDataset(request)
        return CreateMLDatasetJob(self._stub, response.job), MLDataset(self._stub, response.ml_dataset)

    def get(self, name: str):
        """Obtain a specific machine learning dataset (MLDataset).

        Args:
            name: (str) Title of an existing MLDataset.

        Returns:
            MLDataset: Machine Learning Dataset

        Typical example:
            my_feature_view.ml_datasets.get("name")
        """
        request = pb.GetMLDatasetRequest(
            feature_view_id=self._feature_view.id,
            feature_view_version=self._feature_view.version,
            ml_dataset_name=name,
        )

        ml_dataset = self._stub.GetMLDataset(request)
        return MLDataset(self._stub, ml_dataset)

    def list(self):
        """List machine learning datasets (MLDatasets).

        Returns:
            list[MLDataset]: A list of MLDatasets.

        Typical example:
            my_feature_view.ml_datasets.list()
        """
        request = pb.ListMLDatasetsRequest(
            feature_view_id=self._feature_view.id,
            feature_view_version=self._feature_view.version,
        )

        response = self._stub.ListMLDatasets(request)
        return [MLDataset(self._stub, ml_dataset) for ml_dataset in response.ml_datasets]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ai/h2o/featurestore/api/v1/RecommendationProtoApi.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='ai/h2o/featurestore/api/v1/RecommendationProtoApi.proto',
  package='ai.h2o.featurestore.api.v1',
  syntax='proto3',
  serialized_options=b'B\033RecommendationProtoApiProto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n7ai/h2o/featurestore/api/v1/RecommendationProtoApi.proto\x12\x1a\x61i.h2o.featurestore.api.v1\"L\n!RecommendationRegexMatchingPolicy\x12\r\n\x05regex\x18\x01 \x01(\t\x12\x18\n\x10percentage_match\x18\x02 \x01(\r\"\xbf\x01\n\"RecommendationSampleMatchingPolicy\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x01 \x01(\t\x12!\n\x19\x66\x65\x61ture_set_major_version\x18\x02 \x01(\r\x12\x13\n\x0b\x63olumn_name\x18\x03 \x01(\t\x12\x17\n\x0fsample_fraction\x18\x04 \x01(\x01\x12\x16\n\x0e\x66uzzy_distance\x18\x05 \x01(\r\x12\x18\n\x10percentage_match\x18\x06 \x01(\r\"q\n$ListRecommendationClassifierResponse\x12I\n\x0b\x63lassifiers\x18\x01 \x03(\x0b\x32\x34.ai.h2o.featurestore.api.v1.RecommendationClassifier\"\xdd\x01\n\x18RecommendationClassifier\x12\x0c\n\x04name\x18\x01 \x01(\t\x12N\n\x05regex\x18\x02 \x01(\x0b\x32=.ai.h2o.featurestore.api.v1.RecommendationRegexMatchingPolicyH\x00\x12P\n\x06sample\x18\x03 \x01(\x0b\x32>.ai.h2o.featurestore.api.v1.RecommendationSampleMatchingPolicyH\x00\x42\x11\n\x0fmatching_policy\"q\n%CreateRecommendationClassifierRequest\x12H\n\nclassifier\x18\x01 \x01(\x0b\x32\x34.ai.h2o.featurestore.api.v1.RecommendationClassifier\"q\n%UpdateRecommendationClassifierRequest\x12H\n\nclassifier\x18\x01 \x01(\x0b\x32\x34.ai.h2o.featurestore.api.v1.RecommendationClassifier\"@\n%DeleteRecommendationClassifierRequest\x12\x17\n\x0f\x63lassifier_name\x18\x01 \x01(\tB\x1d\x42\x1bRecommendationProtoApiProtob\x06proto3'
)




_RECOMMENDATIONREGEXMATCHINGPOLICY = _descriptor.Descriptor(
  name='RecommendationRegexMatchingPolicy',
  full_name='ai.h2o.featurestore.api.v1.RecommendationRegexMatchingPolicy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='regex', full_name='ai.h2o.featurestore.api.v1.RecommendationRegexMatchingPolicy.regex', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='percentage_match', full_name='ai.h2o.featurestore.api.v1.RecommendationRegexMatchingPolicy.percentage_match', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=87,
  serialized_end=163,
)


_RECOMMENDATIONSAMPLEMATCHINGPOLICY = _descriptor.Descriptor(
  name='RecommendationSampleMatchingPolicy',
  full_name='ai.h2o.featurestore.api.v1.RecommendationSampleMatchingPolicy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_set_id', full_name='ai.h2o.featurestore.api.v1.RecommendationSampleMatchingPolicy.feature_set_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_set_major_version', full_name='ai.h2o.featurestore.api.v1.RecommendationSampleMatchingPolicy.feature_set_major_version', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='column_name', full_name='ai.h2o.featurestore.api.v1.RecommendationSampleMatchingPolicy.column_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_fraction', full_name='ai.h2o.featurestore.api.v1.RecommendationSampleMatchingPolicy.sample_fraction', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fuzzy_distance', full_name='ai.h2o.featurestore.api.v1.RecommendationSampleMatchingPolicy.fuzzy_distance', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='percentage_match', full_name='ai.h2o.featurestore.api.v1.RecommendationSampleMatchingPolicy.percentage_match', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=166,
  serialized_end=357,
)


_LISTRECOMMENDATIONCLASSIFIERRESPONSE = _descriptor.Descriptor(
  name='ListRecommendationClassifierResponse',
  full_name='ai.h2o.featurestore.api.v1.ListRecommendationClassifierResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='classifiers', full_name='ai.h2o.featurestore.api.v1.ListRecommendationClassifierResponse.classifiers', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=359,
  serialized_end=472,
)


_RECOMMENDATIONCLASSIFIER = _descriptor.Descriptor(
  name='RecommendationClassifier',
  full_name='ai.h2o.featurestore.api.v1.RecommendationClassifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='ai.h2o.featurestore.api.v1.RecommendationClassifier.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='regex', full_name='ai.h2o.featurestore.api.v1.RecommendationClassifier.regex', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample', full_name='ai.h2o.featurestore.api.v1.RecommendationClassifier.sample', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='matching_policy', full_name='ai.h2o.featurestore.api.v1.RecommendationClassifier.matching_policy',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=475,
  serialized_end=696,
)


_CREATERECOMMENDATIONCLASSIFIERREQUEST = _descriptor.Descriptor(
  name='CreateRecommendationClassifierRequest',
  full_name='ai.h2o.featurestore.api.v1.CreateRecommendationClassifierRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='classifier', full_name='ai.h2o.featurestore.api.v1.CreateRecommendationClassifierRequest.classifier', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=698,
  serialized_end=811,
)


_UPDATERECOMMENDATIONCLASSIFIERREQUEST = _descriptor.Descriptor(
  name='UpdateRecommendationClassifierRequest',
  full_name='ai.h2o.featurestore.api.v1.UpdateRecommendationClassifierRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='classifier', full_name='ai.h2o.featurestore.api.v1.UpdateRecommendationClassifierRequest.classifier', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=813,
  serialized_end=926,
)


_DELETERECOMMENDATIONCLASSIFIERREQUEST = _descriptor.Descriptor(
  name='DeleteRecommendationClassifierRequest',
  full_name='ai.h2o.featurestore.api.v1.DeleteRecommendationClassifierRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='classifier_name', full_name='ai.h2o.featurestore.api.v1.DeleteRecommendationClassifierRequest.classifier_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=928,
  serialized_end=992,
)

_LISTRECOMMENDATIONCLASSIFIERRESPONSE.fields_by_name['classifiers'].message_type = _RECOMMENDATIONCLASSIFIER
_RECOMMENDATIONCLASSIFIER.fields_by_name['regex'].message_type = _RECOMMENDATIONREGEXMATCHINGPOLICY
_RECOMMENDATIONCLASSIFIER.fields_by_name['sample'].message_type = _RECOMMENDATIONSAMPLEMATCHINGPOLICY
_RECOMMENDATIONCLASSIFIER.oneofs_by_name['matching_policy'].fields.append(
  _RECOMMENDATIONCLASSIFIER.fields_by_name['regex'])
_RECOMMENDATIONCLASSIFIER.fields_by_name['regex'].containing_oneof = _RECOMMENDATIONCLASSIFIER.oneofs_by_name['matching_policy']
_RECOMMENDATIONCLASSIFIER.oneofs_by_name['matching_policy'].fields.append(
  _RECOMMENDATIONCLASSIFIER.fields_by_name['sample'])
_RECOMMENDATIONCLASSIFIER.fields_by_name['sample'].containing_oneof = _RECOMMENDATIONCLASSIFIER.oneofs_by_name['matching_policy']
_CREATERECOMMENDATIONCLASSIFIERREQUEST.fields_by_name['classifier'].message_type = _RECOMMENDATIONCLASSIFIER
_UPDATERECOMMENDATIONCLASSIFIERREQUEST.fields_by_name['classifier'].message_type = _RECOMMENDATIONCLASSIFIER
DESCRIPTOR.message_types_by_name['RecommendationRegexMatchingPolicy'] = _RECOMMENDATIONREGEXMATCHINGPOLICY
DESCRIPTOR.message_types_by_name['RecommendationSampleMatchingPolicy'] = _RECOMMENDATIONSAMPLEMATCHINGPOLICY
DESCRIPTOR.message_types_by_name['ListRecommendationClassifierResponse'] = _LISTRECOMMENDATIONCLASSIFIERRESPONSE
DESCRIPTOR.message_types_by_name['RecommendationClassifier'] = _RECOMMENDATIONCLASSIFIER
DESCRIPTOR.message_types_by_name['CreateRecommendationClassifierRequest'] = _CREATERECOMMENDATIONCLASSIFIERREQUEST
DESCRIPTOR.message_types_by_name['UpdateRecommendationClassifierRequest'] = _UPDATERECOMMENDATIONCLASSIFIERREQUEST
DESCRIPTOR.message_types_by_name['DeleteRecommendationClassifierRequest'] = _DELETERECOMMENDATIONCLASSIFIERREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RecommendationRegexMatchingPolicy = _reflection.GeneratedProtocolMessageType('RecommendationRegexMatchingPolicy', (_message.Message,), {
  'DESCRIPTOR' : _RECOMMENDATIONREGEXMATCHINGPOLICY,
  '__module__' : 'ai.h2o.featurestore.api.v1.RecommendationProtoApi_pb2'
  # @@protoc_insertion_point(class_scope:ai.h2o.featurestore.api.v1.RecommendationRegexMatchingPolicy)
  })
_sym_db.RegisterMessage(RecommendationRegexMatchingPolicy)

RecommendationSampleMatchingPolicy = _reflection.GeneratedProtocolMessageType('RecommendationSampleMatchingPolicy', (_message.Message,), {
  'DESCRIPTOR' : _RECOMMENDATIONSAMPLEMATCHINGPOLICY,
  '__module__' : 'ai.h2o.featurestore.api.v1.RecommendationProtoApi_pb2'
  # @@protoc_insertion_point(class_scope:ai.h2o.featurestore.api.v1.RecommendationSampleMatchingPolicy)
  })
_sym_db.RegisterMessage(RecommendationSampleMatchingPolicy)

ListRecommendationClassifierResponse = _reflection.GeneratedProtocolMessageType('ListRecommendationClassifierResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTRECOMMENDATIONCLASSIFIERRESPONSE,
  '__module__' : 'ai.h2o.featurestore.api.v1.RecommendationProtoApi_pb2'
  # @@protoc_insertion_point(class_scope:ai.h2o.featurestore.api.v1.ListRecommendationClassifierResponse)
  })
_sym_db.RegisterMessage(ListRecommendationClassifierResponse)

RecommendationClassifier = _reflection.GeneratedProtocolMessageType('RecommendationClassifier', (_message.Message,), {
  'DESCRIPTOR' : _RECOMMENDATIONCLASSIFIER,
  '__module__' : 'ai.h2o.featurestore.api.v1.RecommendationProtoApi_pb2'
  # @@protoc_insertion_point(class_scope:ai.h2o.featurestore.api.v1.RecommendationClassifier)
  })
_sym_db.RegisterMessage(RecommendationClassifier)

CreateRecommendationClassifierRequest = _reflection.GeneratedProtocolMessageType('CreateRecommendationClassifierRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATERECOMMENDATIONCLASSIFIERREQUEST,
  '__module__' : 'ai.h2o.featurestore.api.v1.RecommendationProtoApi_pb2'
  # @@protoc_insertion_point(class_scope:ai.h2o.featurestore.api.v1.CreateRecommendationClassifierRequest)
  })
_sym_db.RegisterMessage(CreateRecommendationClassifierRequest)

UpdateRecommendationClassifierRequest = _reflection.GeneratedProtocolMessageType('UpdateRecommendationClassifierRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATERECOMMENDATIONCLASSIFIERREQUEST,
  '__module__' : 'ai.h2o.featurestore.api.v1.RecommendationProtoApi_pb2'
  # @@protoc_insertion_point(class_scope:ai.h2o.featurestore.api.v1.UpdateRecommendationClassifierRequest)
  })
_sym_db.RegisterMessage(UpdateRecommendationClassifierRequest)

DeleteRecommendationClassifierRequest = _reflection.GeneratedProtocolMessageType('DeleteRecommendationClassifierRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETERECOMMENDATIONCLASSIFIERREQUEST,
  '__module__' : 'ai.h2o.featurestore.api.v1.RecommendationProtoApi_pb2'
  # @@protoc_insertion_point(class_scope:ai.h2o.featurestore.api.v1.DeleteRecommendationClassifierRequest)
  })
_sym_db.RegisterMessage(DeleteRecommendationClassifierRequest)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import time

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from ai.h2o.featurestore.api.v1.CoreService_pb2_grpc import CoreServiceStub

from ..entities.scheduled_task import ScheduledTask
from ..job_info import JobInfo


class ScheduledTasks:
    def __init__(self, stub: CoreServiceStub, feature_set):
        self._stub = stub
        self._feature_set = feature_set

    def create_ingest_task(self, request: pb.ScheduleTaskRequest):
        response = self._stub.ScheduleIngestJob(request)
        return ScheduledTask(self._stub, response.task)

    def create_lazy_ingest_task(self, request: pb.ScheduleTaskRequest):
        response = self._stub.ScheduleLazyIngestTask(request)
        return ScheduledTask(self._stub, response.task)

    def tasks(self):
        """List scheduled tasks.

        Returns:
            Generator of scheduled tasks

        Typical example:
            fs.schedule.tasks()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_schedule.html#to-list-scheduled-tasks
        """
        request = pb.ListScheduledTasksRequest()
        request.feature_set_id = self._feature_set.id
        while request:
            response = self._stub.ListScheduledTasks(request)
            if response.next_page_token:
                request.page_token = response.next_page_token
            else:
                request = None
            for task in response.tasks:
                yield ScheduledTask(self._stub, task)

    def get(self, task_id: str):
        """Obtain a scheduled task.

        Args:
            task_id: (str) A unique id of a scheduled task.

        Returns:
            ScheduledTask: An existing scheduled task.

        Typical example:
            task = fs.schedule.get("task_id")
        """
        request = pb.ScheduledTaskId(scheduled_task_id=task_id)
        response = self._stub.GetScheduledTask(request)
        return ScheduledTask(self._stub, response.task)

    def get_lazy_ingest_task(self):
        """Obtain a lazy ingest task.

        This retrieves an existing ingest task which is planned for later ingestion.
        Each major version of a feature set can only contain one lazy ingest task.

        Returns:
            ScheduledTask: A scheduled task.

        Typical example:
            task = fs.schedule.get_lazy_ingest_task()
        """
        request = pb.GetLazyIngestTaskRequest(
            feature_set_id=self._feature_set.id, feature_set_version=self._feature_set.version
        )
        response = self._stub.GetLazyIngestTask(request)
        return ScheduledTask(self._stub, response.task)

    def start_lazy_ingest_task(self):
        """Starts a lazy ingest task."""
        request = pb.LazyIngestRequest(
            feature_set_id=self._feature_set.id, feature_set_version=self._feature_set.version
        )
        response = self._stub.StartLazyIngestTask(request)
        if response.job_id:
            info = JobInfo(self._stub, response.job_id)
            while not self._get_job(response.job_id).done:
                info.show_progress()
                time.sleep(2)
            info.show_progress()  # there is possibility that some progress was pushed before finishing job

    def _get_job(self, job_id):
        return self._stub.GetJob(job_id)



import json

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..retrieve_holder import RetrieveAsLinksCommon
from .base_job import BaseJob


class RetrieveJob(BaseJob, RetrieveAsLinksCommon):
    def _response_method(self, job_id):
        response = self._stub.GetRetrieveAsLinksJobOutput(job_id)
        return response.download_links

    def __repr__(self):
        return json.dumps(
            {
                "job_id": self._job.job_id,
                "job_type": pb.JobType.Name(self._job.job_type),
                "job_done": bool(self._job.done),
                "child_job_ids": self._job.childJobIds,
            }
        )



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from featurestore.core import interactive_console

from ..retrieve_holder import RetrieveHolder
from ..utils import Utils
from .revert_ingest_job import RevertIngestJob


class Ingest:
    def __init__(self, stub, feature_set, ingest):
        self._stub = stub
        self._feature_set = feature_set
        self._ingest = ingest

    def retrieve(self):
        """Retrieve data.

        Returns:
            RetrieveHolder: Returns a link as output for reference.

        Typical example:
            fs.retrieve()
        """
        return RetrieveHolder(self._stub, self._feature_set, "", "", self._ingest.ingest_id)

    @interactive_console.record_stats
    def revert(self):
        """Revert to a specific ingest.

        Reverting creates a new minor version with the data corresponding to the specific ingest removed.

        Typical example:
            ingest.revert()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/ingest_history_api.html#reverting-ingestion
        """
        job = self.revert_async()
        return job.wait_for_result()

    def revert_async(self) -> RevertIngestJob:
        """Create a revert ingestion job for feature set.

        Returns:
            RevertIngestJob: A job for reverting ingestion.

            A job is created with a unique id and type Revert. For example:

            Job(id=<job_id>, type=Revert, done=False, childJobIds=[])

        Typical example:
            ingest.revert_async()

        Raises:
            Exception: Manual revert is not allowed on derived feature set.
        """
        if self._feature_set.derived_from.HasField("transformation"):
            raise Exception("Manual revert is not allowed on derived feature set")

        request = pb.StartRevertIngestJobRequest()
        request.feature_set.CopyFrom(self._feature_set)
        request.ingest_id = self._ingest.ingest_id
        job_id = self._stub.StartRevertIngestJob(request)
        return RevertIngestJob(self._stub, job_id)

    def __repr__(self):
        return Utils.pretty_print_proto(self._ingest)



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from . import interactive_console
from .config import ConfigUtils


class JobInfo:
    def __init__(self, stub, job_id):
        self._stub = stub
        self._job_id = job_id
        self._next_index = 0

    def show_progress(self):
        """Displays job progress in console."""
        if ConfigUtils.is_interactive_print_enabled():
            request = pb.JobProgressInput(job_id=self._job_id.job_id, next_index=self._next_index)
            response = self._stub.GetJobProgress(request)
            if response.progress:
                self._next_index += len(response.progress)
                for progress in response.progress:
                    interactive_console.log(f"Job ID: {self._job_id.job_id}, Status: {progress.message}")

    def get_metrics(self):
        """Obtain the completed metrics of a job.

        Returns:
            dict: A dictionary which consists of completed job progress.

            For example:

            {'Finished setting up spark session.': '6s',
            'Finished reading data from source location to extract schema.': '7s',
            'Schema generation completed.': '0s'}

        Typical example:
            job.get_metrics()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/jobs_api.html?highlight=get_metrics#checking-job-metrics
        """
        if self._stub.GetJob(self._job_id).done:
            request = pb.JobProgressInput(job_id=self._job_id.job_id, next_index=0)
            response = self._stub.GetJobProgress(request)
            if response.progress:
                return {progress.message: f"{progress.duration_in_seconds}s" for progress in response.progress}
        else:
            interactive_console.log("Job is still running. Please wait until the job finishes.")



class AzureKeyCredentials:
    def __init__(self, account_name, account_key):
        self.account_name = account_name
        self.account_key = account_key


class AzureSasCredentials:
    def __init__(self, account_name, sas_token):
        self.account_name = account_name
        self.sas_token = sas_token


class AzurePrincipalCredentials:
    def __init__(self, account_name, client_id, tenant_id, secret):
        self.account_name = account_name
        self.client_id = client_id
        self.tenant_id = tenant_id
        self.secret = secret


class S3Credentials:
    def __init__(self, access_key, secret_key, region=None, endpoint=None, role_arn=None):
        self.access_token = access_key
        self.secret_token = secret_key
        self.region = region
        self.role_arn = role_arn
        self.endpoint = endpoint


class SnowflakeCredentials:
    def __init__(self, user, password):
        self.user = user
        self.password = password


class TeradataCredentials:
    def __init__(self, user, password):
        self.user = user
        self.password = password


class PostgresCredentials:
    def __init__(self, user, password):
        self.user = user
        self.password = password



from logging import config


class FSLogger:
    log_config = {
        "version": 1,
        "root": {"handlers": ["console"], "level": "INFO"},
        "handlers": {"console": {"formatter": "std_out", "class": "logging.StreamHandler", "level": "INFO"}},
        "formatters": {
            "std_out": {
                "format": "%(asctime)s : %(levelname)s : %(module)s : %(message)s",
                "datefmt": "%d-%m-%Y %I:%M:%S",
            }
        },
    }

    config.dictConfig(log_config)



import collections
import logging
import time
from datetime import timedelta
from functools import wraps

from featurestore.core.config import ConfigUtils

METRICS_START_TIME = time.time()
Metrics = collections.namedtuple("Metrics", "func_name sum_time num_calls min_time max_time avg_time")


def log(message):
    if ConfigUtils.is_interactive_print_enabled():
        logging.info(message)


def record_stats(fn):
    @wraps(fn)
    def with_perf(*args, **kwargs):
        start = time.time()
        ret = fn(*args, **kwargs)
        end = time.time()
        delta = timedelta(seconds=end - start).total_seconds()
        logging.info(f"\n\nTime taken - {delta:.3f} seconds")
        return ret

    return with_perf



import datetime
from typing import Optional

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..transformations import SparkPipeline
from ..utils import Utils


class BackfillOption:
    def __init__(
        self,
        from_version: str,
        from_date: Optional[datetime.datetime] = None,
        to_date: Optional[datetime.datetime] = None,
        spark_pipeline: Optional[SparkPipeline] = None,
        feature_mapping: Optional[dict] = None,
    ):
        self._from_version = from_version
        self._from_date = from_date
        self._to_date = to_date
        self._spark_pipeline = spark_pipeline
        self._feature_mapping = feature_mapping

    def _to_proto(self, stub):
        if self._spark_pipeline:
            self._spark_pipeline._initialize(stub)
            spark_pipeline = pb.SparkPipelineTransformation(filename=self._spark_pipeline.pipeline_remote_location)
        else:
            spark_pipeline = None
        return pb.BackfillOptions(
            from_date=Utils.date_time_to_proto_timestamp(self._from_date),
            to_date=Utils.date_time_to_proto_timestamp(self._to_date),
            from_version=self._from_version,
            spark_pipeline=spark_pipeline,
            feature_mapping=self._feature_mapping,
        )



from __future__ import annotations

import json
from typing import TYPE_CHECKING

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
import ai.h2o.featurestore.api.v1.FeatureSetProtoApi_pb2 as FeatureSetApi

from ..utils import Utils
from . import query
from .feature_ref import FeatureRef

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from .feature_set import FeatureSet

CATEGORICAL = FeatureSetApi.FeatureType.Name(FeatureSetApi.FeatureType.CATEGORICAL)
NUMERICAL = FeatureSetApi.FeatureType.Name(FeatureSetApi.FeatureType.NUMERICAL)
TEMPORAL = FeatureSetApi.FeatureType.Name(FeatureSetApi.FeatureType.TEMPORAL)
TEXT = FeatureSetApi.FeatureType.Name(FeatureSetApi.FeatureType.TEXT)
COMPOSITE = FeatureSetApi.FeatureType.Name(FeatureSetApi.FeatureType.COMPOSITE)
AUTOMATIC_DISCOVERY = FeatureSetApi.FeatureType.Name(FeatureSetApi.FeatureType.AUTOMATIC_DISCOVERY)


class Feature:
    def __init__(self, stub, feature_set: FeatureSet, internal_feature, absolute_feature_name):
        self._stub = stub
        self._fs = feature_set
        self._internal_feature = internal_feature
        self._absolute_feature_name = absolute_feature_name

    @property
    def name(self):
        return self._internal_feature.name

    @property
    def version(self):
        return self._internal_feature.version

    @property
    def special(self):
        return self._internal_feature.special

    @special.setter
    def special(self, value):
        update_request = pb.UpdateFeatureRequest(
            feature_set_id=self._fs.id,
            feature_set_version=self._fs.version,
            feature_id=self._internal_feature.id,
            special=value,
            fields_to_update=[pb.FEATURE_SPECIAL],
        )
        response = self._stub.UpdateFeature(update_request)
        self._fs._feature_set = response.updated_feature_set
        self._internal_feature = response.updated_feature

    @property
    def version_change(self):
        return self._internal_feature.version_change

    @property
    def status(self):
        return self._internal_feature.status

    @status.setter
    def status(self, value):
        update_request = pb.UpdateFeatureRequest(
            feature_set_id=self._fs.id,
            feature_set_version=self._fs.version,
            feature_id=self._internal_feature.id,
            status=value,
            fields_to_update=[pb.FEATURE_STATUS],
        )
        response = self._stub.UpdateFeature(update_request)
        self._fs._feature_set = response.updated_feature_set
        self._internal_feature = response.updated_feature

    @property
    def data_type(self):
        return self._internal_feature.data_type

    @property
    def profile(self):
        return FeatureProfile(self._stub, self._fs, self)

    @property
    def description(self):
        return self._internal_feature.description

    @description.setter
    def description(self, value):
        update_request = pb.UpdateFeatureRequest(
            feature_set_id=self._fs.id,
            feature_set_version=self._fs.version,
            feature_id=self._internal_feature.id,
            description=value,
            fields_to_update=[pb.FEATURE_DESCRIPTION],
        )
        response = self._stub.UpdateFeature(update_request)
        self._fs._feature_set = response.updated_feature_set
        self._internal_feature = response.updated_feature

    @property
    def classifiers(self):
        return set(self._internal_feature.classifiers)

    @classifiers.setter
    def classifiers(self, value: set):
        update_request = pb.UpdateFeatureRequest(
            feature_set_id=self._fs.id,
            feature_set_version=self._fs.version,
            feature_id=self._internal_feature.id,
            classifiers=list(value)[:],
            fields_to_update=[pb.FEATURE_CLASSIFIERS],
        )
        response = self._stub.UpdateFeature(update_request)
        self._fs._feature_set = response.updated_feature_set
        self._internal_feature = response.updated_feature

    @property
    def importance(self):
        return self._internal_feature.importance

    @importance.setter
    def importance(self, value):
        update_request = pb.UpdateFeatureRequest(
            feature_set_id=self._fs.id,
            feature_set_version=self._fs.version,
            feature_id=self._internal_feature.id,
            importance=value,
            fields_to_update=[pb.FEATURE_IMPORTANCE],
        )
        response = self._stub.UpdateFeature(update_request)
        self._fs._feature_set = response.updated_feature_set
        self._internal_feature = response.updated_feature

    @property
    def monitoring(self):
        return Monitoring(self._stub, self._fs, self)

    @property
    def special_data(self):
        return FeatureSpecialData(self)

    @property
    def custom_data(self):
        return self._internal_feature.custom_data

    @custom_data.setter
    def custom_data(self, value):
        update_request = pb.UpdateFeatureRequest(
            feature_set_id=self._fs.id,
            feature_set_version=self._fs.version,
            feature_id=self._internal_feature.id,
            custom_data=value,
            fields_to_update=[pb.FEATURE_CUSTOM_DATA],
        )
        response = self._stub.UpdateFeature(update_request)
        self._fs._feature_set = response.updated_feature_set
        self._internal_feature = response.updated_feature

    @property
    def nested_features(self):
        return {
            feature.name: Feature(
                self._stub,
                self._fs,
                feature,
                self._absolute_feature_name + "." + feature.name,
            )
            for feature in self._internal_feature.nested_features
        }

    def _reference(self) -> FeatureRef:
        return FeatureRef(self._absolute_feature_name, self._fs._reference())

    def __lt__(self, value: str) -> query.FilterCondition:
        return query.FilterCondition(self._reference(), query.FilterOperator.LT, value)

    def __le__(self, value: str) -> query.FilterCondition:
        return query.FilterCondition(self._reference(), query.FilterOperator.LE, value)

    def __gt__(self, value: str) -> query.FilterCondition:
        return query.FilterCondition(self._reference(), query.FilterOperator.GT, value)

    def __ge__(self, value: str) -> query.FilterCondition:
        return query.FilterCondition(self._reference(), query.FilterOperator.GE, value)

    def __eq__(self, value: str) -> query.FilterCondition:
        return query.FilterCondition(self._reference(), query.FilterOperator.EQ, value)

    def __ne__(self, value: str) -> query.FilterCondition:
        return query.FilterCondition(self._reference(), query.FilterOperator.NE, value)

    def __repr__(self):
        return Utils.pretty_print_proto(self._internal_feature)

    def __str__(self):
        return (
            f"Feature name      : {self.name} \n"
            f"Description       : {self.description} \n"
            f"Version           : {self.version} \n"
            f"Data type         : {self.data_type} \n"
            f"Sensitive         : {self.special_data.sensitive} \n"
            f"Nested features   : {self._custom_feature_fields()} \n"
            f"Classifiers       : {self.classifiers} \n"
        )

    def _custom_feature_fields(self):
        s = dict()
        for feature in self.nested_features:
            s.update({feature: self.nested_features.get(feature).data_type})
        return json.dumps(s, indent=5)


class FeatureProfile:
    def __init__(self, stub, feature_set, feature):
        self._stub = stub
        self._fs = feature_set
        self._feature = feature

    @property
    def feature_type(self):
        return FeatureSetApi.FeatureType.Name(self._feature._internal_feature.profile.feature_type)

    @feature_type.setter
    def feature_type(self, value):
        valid_values = FeatureSetApi.FeatureType.keys()
        if value.upper() in valid_values:
            update_request = pb.UpdateFeatureRequest(
                feature_set_id=self._fs.id,
                feature_set_version=self._fs.version,
                feature_id=self._feature._internal_feature.id,
                type=FeatureSetApi.FeatureType.Value(value.upper()),
                fields_to_update=[pb.FEATURE_TYPE],
            )
            response = self._stub.UpdateFeature(update_request)
            self._feature._fs._feature_set = response.updated_feature_set
            self._feature._internal_feature = response.updated_feature
        else:
            raise Exception("Invalid feature type. Supported values are: " + ", ".join(map(str, valid_values)))

    @property
    def categorical_statistics(self):
        return CategoricalStatistics(self._feature._internal_feature.profile)

    @property
    def statistics(self):
        return FeatureStatistics(self._feature._internal_feature.profile)

    def __repr__(self):
        return Utils.pretty_print_proto(self._feature._internal_feature.profile)

    def __str__(self):
        return (
            f"Feature type  : {self.feature_type} \n"
            f"Categorical     \n{Utils.output_indent_spacing(str(self.categorical_statistics), '    ')}"
            f"Statistics      \n{Utils.output_indent_spacing(str(self.statistics), '  ')}"
        )


class FeatureStatistics:
    def __init__(self, feature):
        self._feature = feature
        self._stats = self._feature.statistics

    @property
    def max(self):
        return self._stats.max

    @property
    def mean(self):
        return self._stats.mean

    @property
    def median(self):
        return self._stats.median

    @property
    def min(self):
        return self._stats.min

    @property
    def stddev(self):
        return self._stats.stddev

    @property
    def stddev_rec_count(self):
        return self._stats.stddev_rec_count

    @property
    def null_count(self):
        return self._stats.null_count

    @property
    def nan_count(self):
        return self._stats.nan_count

    @property
    def unique(self):
        return self._stats.unique

    def __repr__(self):
        return Utils.pretty_print_proto(self._stats)

    def __str__(self):
        return (
            f"Max                : {self.max} \n"
            f"Mean               : {self.mean} \n"
            f"Median             : {self.median} \n"
            f"Min                : {self.min} \n"
            f"Standard deviation : {self.stddev} \n"
            f"Records count      : {self.stddev_rec_count} \n"
            f"Null values count  : {self.null_count} \n"
            f"NaN values count   : {self.nan_count} \n"
            f"Unique             : {self.unique} \n"
        )


class CategoricalStatistics:
    def __init__(self, feature):
        self._feature = feature
        self._categorical = self._feature.categorical

    @property
    def unique(self):
        return self._categorical.unique

    @property
    def top(self):
        return [FeatureTop(top) for top in self._categorical.top]

    def __repr__(self):
        return Utils.pretty_print_proto(self._categorical)

    def __str__(self):
        return f"Unique    : {self.unique} \n" f"Top       : {self.top} \n"


class FeatureTop:
    def __init__(self, feature):
        self._feature = feature
        self._top = self._feature.top

    @property
    def name(self):
        return self._top.name

    @property
    def count(self):
        return self._top.count

    def __repr__(self):
        return Utils.pretty_print_proto(self._top)

    def __str__(self):
        return f"Name  : {self.name} \n" f"Count : {self.count} \n"


class Monitoring:
    def __init__(self, stub, feature_set, feature):
        self._stub = stub
        self._fs = feature_set
        self._feature = feature

    @property
    def anomaly_detection(self):
        return self._feature._internal_feature.monitoring.anomaly_detection

    @anomaly_detection.setter
    def anomaly_detection(self, value):
        update_request = pb.UpdateFeatureRequest(
            feature_set_id=self._fs.id,
            feature_set_version=self._fs.version,
            feature_id=self._feature._internal_feature.id,
            anomaly_detection=value,
            fields_to_update=[pb.FEATURE_ANOMALY_DETECTION],
        )
        response = self._stub.UpdateFeature(update_request)
        self._feature._fs._feature_set = response.updated_feature_set
        self._feature._internal_feature = response.updated_feature

    def __repr__(self):
        return Utils.pretty_print_proto(self._feature._internal_feature.monitoring)

    def __str__(self):
        return f"Anomaly detection : {self.anomaly_detection} \n"


class FeatureSpecialData:
    def __init__(self, feature):
        self._feature = feature

    @property
    def spi(self):
        return self._feature._internal_feature.special_data.spi

    @property
    def pci(self):
        return self._feature._internal_feature.special_data.pci

    @property
    def rpi(self):
        return self._feature._internal_feature.special_data.rpi

    @property
    def demographic(self):
        return self._feature._internal_feature.special_data.demographic

    @property
    def sensitive(self):
        return self._feature._internal_feature.special_data.sensitive

    def __repr__(self):
        return Utils.pretty_print_proto(self._feature._internal_feature.special_data)

    def __str__(self):
        if self.sensitive:
            return (
                f"spi          : {self.spi} \n"
                f"pci          : {self.pci} \n"
                f"rpi          : {self.rpi} \n"
                f"demographic  : {self.demographic} \n"
                f"sensitive    : {self.sensitive} \n"
            )
        return (
            f"spi          : *** \n"
            f"pci          : *** \n"
            f"rpi          : *** \n"
            f"demographic  : *** \n"
            f"sensitive    : {self.sensitive} \n"
        )



import asyncio
import datetime
import inspect
import logging
import re
import sys
import webbrowser

import grpc
from google.protobuf.empty_pb2 import Empty

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from featurestore.core.config import ConfigUtils

from .collections.pats import PersonalAccessTokens


class AuthWrapper:
    ACCESS_TOKEN_EXPIRES_SOON_SECS = 60

    def __init__(self, stub):
        self._stub = stub
        self._access_token = None
        self._access_token_expiration_date = None
        self._get_access_token_external = None
        self._props = ConfigUtils.collect_properties()
        self.pats = PersonalAccessTokens(self._stub)

    def get_active_user(self):
        """Obtain currently active user details.

        Returns:
            UserBasicInfo: Logged-in user details.

            For example:

            id: <user_id>
            name: <user_name>
            email: <user_email>
        """
        request = Empty()
        return self._stub.GetActiveUser(request).user

    def set_obtain_access_token_method(self, method):
        """Set a valid user access token.

        This will obtain an access token from the external environment to use for authentication.

        Args:
            method: (Callable) A method that returns a token from the external environment.

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/authentication.html#authentication-via-access-token-from-external-environment
        """
        self._get_access_token_external = method

    def logout(self):
        """Logout the current user session."""
        is_success, token_or_error = self._obtain_token()
        if is_success:
            if not AuthWrapper._is_personal_access_token(self._props["token"].data):
                request = pb.LogoutRequest()
                request.refresh_token = self._props["token"].data
                self._stub.Logout(request)
                self._access_token = None
                self._access_token_expiration_date = None
            ConfigUtils.delete_property(self._props, ConfigUtils.TOKEN_KEY)
            logging.info("You have been logged out.")
        else:
            logging.info("You are not logged in.")

    def login(self, open_browser=True):
        """User login via identity provider.

        This opens the returned URL in the browser (if it's not possible to open the browser, you will
        have to do this manually) and wait for the refresh token. Returned refresh tokens will be
        saved into the client’s configuration file.

        Args:
            open_browser: (bool) If True, opens the URL in the browser. Defaults to True.

        Raises:
            AuthException: Incorrect response or if user doesn't log in via browser,
              logging session will be expired.

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/authentication.html#authentication-via-refresh-token-from-identity-provider
        """
        try:
            sys.tracebacklimit = None
            for response in self._stub.Login(Empty()):
                if response.HasField("login_url"):
                    if open_browser:
                        try:
                            webbrowser.get()
                            webbrowser.open(response.login_url)
                            logging.info(f"Opening browser to visit: {response.login_url}")
                        except webbrowser.Error:
                            logging.error(
                                f"Browser is not supported: Please visit "
                                f"{response.login_url} to continue authentication."
                            )
                    else:
                        logging.info(f"Please visit {response.login_url} to continue authentication.")
                elif response.HasField("refresh_token"):
                    self.set_auth_token(response.refresh_token)
                else:
                    sys.tracebacklimit = 0
                    raise AuthException("Incorrect response")
        except grpc._channel._MultiThreadedRendezvous:
            logging.error("Your logging session expired, please login again by running client.auth.login().")

    def set_auth_token(self, token):
        """Set user token for authentication.

        Args:
            token: (str) A personal access token.

        Raises:
            AuthException: If user is not authenticated or token is no longer valid.
        """
        sys.tracebacklimit = None
        ConfigUtils.store_token(self._props, token)
        if not AuthWrapper._is_personal_access_token(token):
            self._access_token = None
            self._access_token_expiration_date = None
            is_success, token_or_error = self._obtain_token()
            if not is_success:
                sys.tracebacklimit = 0
                raise AuthException(token_or_error)

    @staticmethod
    def _is_personal_access_token(token: str) -> bool:
        return re.match(r"^[a-z0-9]{3}_.*", token)

    def _is_access_token_expired(self):
        if not self._access_token:
            return True
        if self._access_token_expiration_date is not None:
            expires_in = (self._access_token_expiration_date - datetime.datetime.now()).total_seconds()
            return expires_in <= AuthWrapper.ACCESS_TOKEN_EXPIRES_SOON_SECS

        return False

    def _obtain_token(self):
        if self._get_access_token_external is not None:
            external_token = self._get_access_token_external()
            if inspect.isawaitable(external_token):
                loop = asyncio.get_event_loop()
                return True, loop.run_until_complete(external_token)
            else:
                return True, external_token
        elif ConfigUtils.TOKEN_KEY not in self._props:
            return (
                False,
                "You are not authenticated. Set personal access token or execute client.auth.login() method",
            )
        elif AuthWrapper._is_personal_access_token(ConfigUtils.get_token(self._props)):
            return True, ConfigUtils.get_token(self._props)
        elif self._is_access_token_expired():
            request = pb.RefreshTokenRequest()
            request.refresh_token = ConfigUtils.get_token(self._props)
            try:
                resp = self._stub.GetAccessToken(request)
            except grpc.RpcError:
                return (
                    False,
                    "The authentication token is no longer valid. Please login again.",
                )

            self._access_token = resp.access_token
            ConfigUtils.store_token(self._props, resp.refresh_token)
            self._access_token_expiration_date = datetime.datetime.now() + datetime.timedelta(seconds=resp.expires_in)
            return True, self._access_token
        else:
            return True, self._access_token

    def __repr__(self):
        return "This class wraps together methods related to Authentication"


class AuthException(Exception):
    pass
